using RiskAdjustedLinearizations, SparseArrays, SparseDiffTools, Test
include(joinpath(dirname(@__FILE__), "..", "..", "examples", "rbc_cc", "rbc_cc.jl"))
include(joinpath(dirname(@__FILE__), "..", "..", "examples", "crw", "crw.jl"))

# Set up
n_strips = 0
m_rbc_cc = RBCCampbellCochraneHabits()
m_crw = CoeurdacierReyWinant()

# Test sparse Jacobians on RBC-CC
m_dense = rbc_cc(m_rbc_cc, n_strips)
z0 = copy(m_dense.z)
y0 = copy(m_dense.y)
Ψ0 = copy(m_dense.Ψ)

m = rbc_cc(m_rbc_cc, n_strips; sparse_jacobian = [:μ, :ξ, :𝒱])
@testset "Construct a RiskAdjustedLinearization that exploits sparsity in Jacobians (using RBC-CC)" begin
    @test isempty(m_dense.linearization.sparse_jac_caches)
    @test m.z ≈ z0
    @test m.y ≈ y0
    @test m.Ψ ≈ Ψ0
    for k in [:μz, :μy, :ξz, :ξy, :J𝒱]
        @test haskey(m.linearization.sparse_jac_caches, k)
        if k != :J𝒱
            @test issparse(m.linearization.sparse_jac_caches[k][:sparsity])
            @test isa(m.linearization.sparse_jac_caches[k][:colorvec], AbstractVector{Int})
        end
    end
    @test m.linearization.sparse_jac_caches[:J𝒱][:colorvec] == 1:2
    @test m.linearization.sparse_jac_caches[:J𝒱][:sparsity] == ones(size(m.Ψ))
end

@testset "Update a RiskAdjustedLinearization to exploit sparsity in Jacobians (using RBC-CC)" begin
    update_sparsity_pattern!(m_dense, :𝒱)
    for k in [:μz, :μy, :ξz, :ξy]
        @test !haskey(m_dense.linearization.sparse_jac_caches, k)
    end
    @test m_dense.linearization.sparse_jac_caches[:J𝒱][:colorvec] == 1:2
    @test m_dense.linearization.sparse_jac_caches[:J𝒱][:sparsity] == ones(size(m.Ψ))

    update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱])
    for k in [:μz, :μy, :ξz, :ξy]
        @test haskey(m_dense.linearization.sparse_jac_caches, k)
        @test issparse(m_dense.linearization.sparse_jac_caches[k][:sparsity])
        @test isa(m_dense.linearization.sparse_jac_caches[k][:colorvec], AbstractVector{Int})
    end

    @test_broken update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱]; sparsity_detection = true)
end

@testset "Calculate risk-adjusted linearization with sparse autodiff (using RBC-CC)" begin
    # Now provide the sparsity pattern and matrix coloring vector
    # to update the Jacobians of objects
    rbc_cc_out = JLD2.jldopen(joinpath(dirname(@__FILE__), "..", "reference", "rbccc_sss_iterative_output.jld2"), "r")
    m_dense = rbc_cc(m_rbc_cc, n_strips) # recompute to get dense Jacobians again
    update!(m_dense, vec(rbc_cc_out["z_rss"]), vec(rbc_cc_out["y_rss"]), rbc_cc_out["Psi_rss"])
    ztrue = copy(m_dense.z)
    ytrue = copy(m_dense.y)
    Ψtrue = copy(m_dense.Ψ)

    sparsity = Dict{Symbol, SparseMatrixCSC{Float64, Int64}}()
    colorvec = Dict{Symbol, Vector{Int64}}()
    sparsity[:μz] = sparse(m_dense[:Γ₁])
    sparsity[:μy] = sparse(m_dense[:Γ₂])
    sparsity[:ξz] = sparse(m_dense[:Γ₃])
    sparsity[:ξy] = sparse(m_dense[:Γ₄])
    sparsity[:J𝒱] = sparse(m_dense[:JV])
    for (k, v) in sparsity
        colorvec[k] = matrix_colors(v)
    end

    # Check updating dense Jacobians works
    update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱])
    try # prone to weird non-deterministic behavior in nlsolve
        solve!(m_dense, ztrue * 1.005, ytrue * 1.005, Ψtrue * 1.005; algorithm = :relaxation,
               ftol = 1e-6, tol = 1e-5, verbose = :none)
        @test norm(steady_state_errors(m_dense), Inf) < 1e-4
    catch e
        println("Updating dense Jacobian with sparse Jacobian methods did not pass")
    end
    @test_broken update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱]; sparsity_detection = true)

    # Check updating sparse Jacobians w/new patterns works
    update_sparsity_pattern!(m, :𝒱; sparsity = sparsity,
                             colorvec = colorvec)
    try # prone to weird non-deterministic behavior in nlsolve
        solve!(m, ztrue * 1.005, ytrue * 1.005, Ψtrue * 1.005; algorithm = :relaxation,
               ftol = 1e-6, tol = 1e-5, verbose = :none)
        @test norm(steady_state_errors(m), Inf) < 1e-4
    catch e
        println("Updating sparsity pattern of 𝒱 for an RAL w/sparse methods did not pass")
    end

    update_sparsity_pattern!(m, [:μ, :ξ, :𝒱]; sparsity = sparsity)
    update_sparsity_pattern!(m, [:μ, :ξ, :𝒱]; sparsity = sparsity,
                             colorvec = colorvec)
    try # prone to weird non-deterministic behavior in nlsolve
        solve!(m, ztrue * 1.005, ytrue * 1.005, Ψtrue * 1.005; algorithm = :relaxation,
               ftol = 1e-6, tol = 1e-5, verbose = :none)
        @test norm(steady_state_errors(m), Inf) < 1e-4
    catch e
        println("Updating sparsity pattern of μ, ξ, and 𝒱 for an RAL w/sparse methods did not pass")
    end

    close(rbc_cc_out)
    # caching appears to be failing somehow; the caches of μ, ξ, and 𝒱 are being set to NaN unexpectedly
    @test_broken solve!(m, ztrue * 1.005, ytrue * 1.005, Ψtrue * 1.005; algorithm = :homotopy, verbose = :none)
    #=
    @test m.z ≈ m_dense.z atol=1e-6
    @test m.y ≈ m_dense.y atol=1e-6
    @test m.Ψ ≈ m_dense.Ψ atol=1e-6
    =#
end

# Test sparse Jacobians on CRW
m_dense = crw(m_crw)
z0 = copy(m_dense.z)
y0 = copy(m_dense.y)
Ψ0 = copy(m_dense.Ψ)

m = crw(m_crw; Ψ = zero(Ψ0), sparse_jacobian = [:μ, :ξ, :𝒱])
m_dense.Ψ .= 0.
@testset "Construct a RiskAdjustedLinearization that exploits sparsity in Jacobians (using CRW)" begin
    @test isempty(m_dense.linearization.sparse_jac_caches)
    @test m.z ≈ z0
    @test m.y ≈ y0
    for k in [:μz, :μy, :ξz, :ξy, :J𝒱]
        @test haskey(m.linearization.sparse_jac_caches, k)
        if k != :J𝒱
            @test issparse(m.linearization.sparse_jac_caches[k][:sparsity])
            @test isa(m.linearization.sparse_jac_caches[k][:colorvec], AbstractVector{Int})
        end
    end
    @test m.linearization.sparse_jac_caches[:J𝒱][:colorvec] == 1:3
    @test m.linearization.sparse_jac_caches[:J𝒱][:sparsity] == ones(size(m.Ψ))
end

@testset "Update a RiskAdjustedLinearization to exploit sparsity in Jacobians (using CRW)" begin
    update_sparsity_pattern!(m_dense, :𝒱)
    for k in [:μz, :μy, :ξz, :ξy]
        @test !haskey(m_dense.linearization.sparse_jac_caches, k)
    end
    @test m_dense.linearization.sparse_jac_caches[:J𝒱][:colorvec] == 1:3
    @test m_dense.linearization.sparse_jac_caches[:J𝒱][:sparsity] == ones(size(m.Ψ))

    update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱])
    for k in [:μz, :μy]
        @test haskey(m_dense.linearization.sparse_jac_caches, k)
        @test issparse(m_dense.linearization.sparse_jac_caches[k][:sparsity])
        @test isa(m_dense.linearization.sparse_jac_caches[k][:colorvec], AbstractVector{Int})
    end
    @test_broken update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱]; sparsity_detection = true)
end

@testset "Calculate risk-adjusted linearization with sparse autodiff (using CRW)" begin
    # Now provide the sparsity pattern and matrix coloring vector
    # to update the Jacobians of objects
    m_dense = crw(m_crw) # recompute to get dense Jacobians again
    crw_out = JLD2.jldopen(joinpath(dirname(@__FILE__), "..", "reference/crw_sss.jld2"), "r")
    update!(m_dense, vec(crw_out["z_rss"]), vec(crw_out["y_rss"]), copy(crw_out["Psi_rss"]))
    ztrue = copy(m_dense.z)
    ytrue = copy(m_dense.y)
    Ψtrue = copy(m_dense.Ψ)

    sparsity = Dict{Symbol, SparseMatrixCSC{Float64, Int64}}()
    colorvec = Dict{Symbol, Vector{Int64}}()
    sparsity[:μz] = sparse(m_dense[:Γ₁])
    sparsity[:μy] = sparse(m_dense[:Γ₂])
    sparsity[:ξz] = sparse(ones(size(m_dense[:Γ₃])))
    sparsity[:ξy] = sparse(m_dense[:Γ₄])
    sparsity[:J𝒱] = sparse(m_dense[:JV])
    for (k, v) in sparsity
        if k == :ξz
            colorvec[k] = 1:3
        elseif k ==  :J𝒱
            colorvec[k] = ones(Int64, 3)
        else
            colorvec[k] = matrix_colors(v)
        end
    end

    # Check updating dense Jacobians works
    update_sparsity_pattern!(m_dense, [:μ, :ξ])
    try
        solve!(m_dense, ztrue, ytrue, Ψtrue; algorithm = :relaxation, ftol = 5e-4, tol = 1e-3, verbose = :none)
        @test norm(steady_state_errors(m_dense), Inf) < 1e-3
    catch e
        println("Updating dense Jacobian with sparse Jacobian methods did not pass")
    end

    # Check updating sparse Jacobians w/new patterns works
    update_sparsity_pattern!(m, :𝒱; sparsity = sparsity,
                             colorvec = colorvec)
    try
        solve!(m, ztrue, ytrue, Ψtrue; algorithm = :relaxation, ftol = 5e-4, tol = 1e-3, verbose = :none)
        @test norm(steady_state_errors(m), Inf) < 1e-3
    catch e
        println("Updating sparsity pattern of 𝒱 for an RAL w/sparse methods did not pass")
    end

    update_sparsity_pattern!(m, [:μ, :ξ]; sparsity = sparsity) # don't do for 𝒱 b/c its Jacobian is all zeros
    update_sparsity_pattern!(m, [:μ, :ξ, :𝒱]; sparsity = sparsity,
                             colorvec = colorvec)
    try
        solve!(m, ztrue, ytrue, Ψtrue; algorithm = :relaxation, ftol = 5e-4, tol = 1e-3, verbose = :none)
        @test norm(steady_state_errors(m), Inf) < 1e-3
    catch e
        println("Updating sparsity pattern of μ, ξ, and 𝒱 for an RAL w/sparse methods did not pass")
    end
    @test_broken update_sparsity_pattern!(m_dense, [:μ, :ξ, :𝒱]; sparsity_detection = true)

    @test_broken solve!(m, ztrue, ytrue, Ψtrue; algorithm = :homotopy, ftol = 5e-4, tol = 1e-3, verbose = :none)
end
